/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Felix Weiglhofer [committer], Dominik Smith */
#include "Exceptions.h"
#include "ReadoutPars.h"

using namespace cbm::algo::tof::config;

ReadoutPars ReadoutPars::MakeMCBM2022()
{
  // This here refers to the mCBM 2022 setup.
  // Taken from CbmMcbm2018TofPar in combination with macro/beamtime/mcbm2022/mTofCriParNickel.par
  // and mTofCriParNickel_withBmon.par (?)

  ReadoutPars pars;

  pars.nFebsPerComponent = 10;
  pars.nAsicsPerFeb      = 8;
  pars.nChannelsPerAsic  = 4;
  pars.nCrobPerComponent = 2;

  constexpr int NumCrob = 16;

  // Module Identifier connected to Gbtx link, has to match geometry
  const int32_t moduleId[NumCrob] = {0, 0, 1, 1, 0, 0, 2, 2, 3, 3, 4, 4, 0, -1, 0, 0};

  // type of Rpcs connected to Gbtx link
  const int32_t rpcType[NumCrob] = {0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 9, 9, 6, 9};

  // side of Rpcs connected to Gbtx link, i.e. 0 or 1
  const int32_t rpcSide[NumCrob] = {1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 8, 0};

  // number of Rpcs connected to Gbtx link, i.e. 3 or 5
  const int32_t numRpc[NumCrob] = {5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5};

  for (int i = 0; i < NumCrob; ++i) {
    pars.crobs.push_back({moduleId[i], rpcType[i], rpcSide[i], numRpc[i]});
  }

  pars.eqIds = {0xabc0, 0xabc1, 0xabc2, 0xabc3, 0xabc4, 0xabc5, 0xabc6, 0xabc7};

  pars.elink2Asic = {27, 2,  7,  3,  31, 26, 30, 1,  33, 37, 32, 13, 9,  14, 10, 15, 17, 21, 16, 35,
                     34, 38, 25, 24, 0,  6,  20, 23, 18, 22, 28, 4,  29, 5,  19, 36, 39, 8,  12, 11};

  // Mapping in Readout chain PCBs
  pars.asic2PadI = {
    3,  2,  1,  0,  7,  6,  5,  4,  11, 10, 9,  8,  15, 14, 13, 12,
    19, 18, 17, 16, 23, 22, 21, 20, 27, 26, 25, 24, 31, 30, 29, 28};  //! Map from GET4 channel to PADI channel

  if (pars.NElinksPerCrob() != static_cast<i32>(pars.elink2Asic.size())) {
    throw FatalError("tof::ReadoutPars::MakeMCBM2022 => size of elink2Asic not matching, {} vs {}, stop there!",
                     static_cast<uint32_t>(pars.elink2Asic.size()), static_cast<uint32_t>(pars.NElinksPerComponent()));
  }
  if (pars.NChansPerFeb() != static_cast<i32>(pars.asic2PadI.size())) {
    throw FatalError("tof::ReadoutPars::MakeMCBM2022 => size of asic2PadI not matching, {} vs {}, stop there!",
                     static_cast<uint32_t>(pars.elink2Asic.size()), static_cast<uint32_t>(pars.NChansPerFeb()));
  }

  return pars;
}
